# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:52 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : parallel.py
# @Desc   : for Parallelization

import joblib
from inspect import signature


def job_factory(
    func, feeder, parallel: joblib.Parallel = None, **kwargs,
):
    """
    parallel run func with data provided by feeder

    :param func:
    :param feeder:
    :param parallel:
    :param kwargs:
    :return:
    """
    kwargs = {
        k: v
        for k, v in kwargs.items()
        if k in signature(joblib.Parallel).parameters.keys()
    }
    if parallel is None:
        # hack on n_jobs to avoid unnecessary overhead
        n_jobs = kwargs.get("n_jobs", -1)
        if n_jobs < 0:
            n_jobs = joblib.cpu_count() + n_jobs + 1

        # retricted by total
        try:
            total = feeder.total
            n_jobs = min(n_jobs, total)
        except:
            pass

        kwargs["n_jobs"] = n_jobs

        # shortcut when 1
        if n_jobs == 1:
            return [func(i) for i in feeder]

        parallel = joblib.Parallel(**kwargs)
    return parallel(joblib.delayed(func)(i) for i in feeder)


# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:53 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : overlap.py
# @Desc   : overlapping indicators


import pandas as pd
import numpy as np
from inspect import isfunction


def ohlc4(df, **kwargs):
    s_val = df[["open", "high", "low", "close"]].mean(axis=1)
    s_val.name = "ohlc4"
    return s_val


def val_max_x(df, x, val="close", name_only=False, **kwargs):
    name = f"{val}_max{x}"
    if name_only:
        return name

    s_val = df[val].rolling(int(x)).max()
    s_val.name = name
    return s_val


def val_min_x(df, x, val="close", name_only=False, **kwargs):
    name = f"{val}_min{x}"
    if name_only:
        return name

    s_val = df[val].rolling(int(x)).min()
    s_val.name = name
    return s_val


def rx(df, x, col="close", name_only=False, **kwargs):
    """Return with window x: 将当前的col与 x天前的col比较，计算出的return。支持不同的col。"""
    name = f"r{x}" if col == "close" else f"r{x}_{col}"
    if name_only:
        return name

    s_val = df[col] / df[col].shift(x) - 1
    s_val.name = name
    return s_val


def lrx(df, x, name_only=False, **kwargs):
    """Lower Return with window x: 将当前的close与 max[x天前的close, x-1天前的open]比较，计算出的return"""
    name = f"lr{x}"
    if name_only:
        return name

    dft = df[["open", "close"]].copy()
    dft["next_open"] = dft.open.shift(-1)
    dft["high_ref"] = dft[["close", "next_open"]].max(axis=1)

    s_val = dft["close"] / dft["high_ref"].shift(x) - 1
    s_val.name = name
    return s_val


def grx(df, x, name_only=False, **kwargs):
    """Greater Return with window x: 将当前的close与 min[x天前的close, x-1天前的open]比较，计算出的return"""
    name = f"gr{x}"
    if name_only:
        return name

    dft = df[["open", "close"]].copy()
    dft["next_open"] = dft.open.shift(-1)
    dft["low_ref"] = dft[["close", "next_open"]].min(axis=1)

    s_val = dft["close"] / dft["low_ref"].shift(x) - 1
    s_val.name = name
    return s_val


def volma_x(df, x, name_only=False, **kwargs):
    name = f"volma{x}"
    if name_only:
        return name

    s_val = df["volume"].rolling(x).mean()
    s_val.name = name
    return s_val


def volma_dx(df, x, name_only=False, **kwargs):
    name = f"volma_d{x}"
    if name_only:
        return name

    s_volma_dx = df.loc[df.close < df.close.shift(), "volume"].rolling(x).mean()
    s_val = s_volma_dx.reindex(index=df.index).fillna(method="ffill")
    s_val.name = name
    return s_val


def volma_ux(df, x, name_only=False, **kwargs):
    """volume moving average of recent x up days"""
    name = f"volma_u{x}"
    if name_only:
        return name

    s_volma_ux = df.loc[df.close > df.close.shift(), "volume"].rolling(x).mean()
    s_val = s_volma_ux.reindex(index=df.index).fillna(method="ffill")
    s_val.name = name
    return s_val


def qrr_x(df, x, name_only=False, **kwargs):
    name = f"qrr_{x}"
    if name_only:
        return name

    s_val = df.volume / df.volume.rolling(x).mean().shift(-1)
    s_val.name = name
    return s_val


# ----keep at the end
_group = {
    k: v for k, v in locals().items() if isfunction(v) and v.__module__ == __name__
}



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:50 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : __init__.py.py
# @Desc   : lutar's init file

from .commons import *
from .techtable import TechTable

TT = TechTable


# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:51 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : helpers.py
# @Desc   : useful funcs and decorators

import warnings
from functools import wraps
import inspect
import pandas as pd

from .parallel import job_factory


def resolve_irange(
    istart, iend, i_max=None, i_min=0, start=None, end=None, dt_index=None
):
    """
    Decide the real istart and iend
    """
    if i_max is None:
        i_max = dt_index.size - 1

    istart_list = []
    if istart:
        if istart < 0:
            istart = i_max + 1 + istart
        istart_list.append(istart)

    if start:
        try:
            real_start = dt_index[dt_index >= start][0]
            istart2 = dt_index.get_loc(real_start)
        except IndexError:
            # cannot find real_start since no days >= start
            # no data to calculate
            istart2 = i_max + 1
        except TypeError:
            # cannot compare
            # no constraint induced
            istart2 = 0
        istart_list.append(istart2)

    # use the earlier istart
    istart = max(i_min, min(istart_list) if istart_list else 0)

    iend_list = []
    if iend:
        if iend < 0:
            iend = i_max + 1 + iend
        iend_list.append(iend)

    if end:
        try:
            real_end = dt_index[dt_index <= end][-1]
            iend2 = dt_index.get_loc(real_end)
        except IndexError:
            iend2 = i_min - 1
        except TypeError:
            iend2 = i_max
        iend_list.append(iend2)

    iend = min(i_max, max(iend_list) if iend_list else i_max)
    return istart, iend


# -------- warnings
def custom_formatwarning(msg, *args, **kwargs):
    # ignore everything except the message
    return f"{msg}\n"


warnings.formatwarning = custom_formatwarning
# warnings.simplefilter(action="ignore", category=FutureWarning)


def missing_warning(cols_needed, df_columns, ti_name, msg=""):
    missing_list = []
    for c in cols_needed:
        if c not in df_columns:
            missing_list.append(c)
    if missing_list:
        value_missing = f"Fields: {missing_list} cannot be found. Needed by {ti_name.upper()}. \n{msg}\n"
        warnings.warn(value_missing)
        return True

    return False


# -------- warnings

# -------- decorators
def magicdf(func):
    """
    # This decorator enables the following functions:

    * Accept TechTable so that all funcs can be directly accessed
    * Enable 'name_only'/'fresh'/'save' for those func without the parameter
    * Return the result directly if it already exists as a column

    # Parameters

    * name_only: only return the unique name of the generated series
    * fresh: force re-calculate values by the func
    * save: save into the original dataframe

    # standard format of func

    ```
    def func(df, x, param1="aaa", param2='bbb', name_only=False, **kwargs):
        name = f"r{x}"
        if name_only:
            return name


        s_val = ... # main logic
        s_val.name = name
        return s_val
    ```

    """

    @wraps(func)
    def wrapper(df, *args, **kwargs):
        # shortcut for gen funcs
        if func.__name__.startswith("lhgen_"):
            return func(df=df, *args, **kwargs)

        func_sig = inspect.signature(func)
        trans_inputs = func_sig.bind(df, *args, **kwargs)
        trans_inputs.apply_defaults()
        input_dict = {}
        for k, v in trans_inputs.arguments.items():
            if k != "kwargs":
                input_dict[k] = v
            else:
                input_dict.update(v)

        # auto-save
        # 1. get the name
        flag_name_only = input_dict.get("name_only", False)  # 临时保存
        if "name_only" not in func_sig.parameters:  # 名字固定的TI，无需设置name_only参数
            ti_name = func.__name__
        else:
            input_dict["name_only"] = True
            ti_name = func(**input_dict)

        if flag_name_only:
            return ti_name

        # 2. get the value
        if ti_name in df.columns and not input_dict.get("fresh", False):
            return df[ti_name]

        # re-calculate
        input_dict["name_only"] = False
        result = func(**input_dict)

        # 3. save
        if input_dict.get("save", True) and isinstance(result, pd.Series):
            df[ti_name] = result
        return result

    return wrapper


# -------- decorators

# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:53 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : __init__.py.py
# @Desc   : The funcs package's init file


import os
from inspect import isfunction
from lutar.commons import magicdf

func_dict = {}

for f in os.listdir(os.path.dirname(os.path.abspath(__file__))):
    if f.endswith(".py") and f != "__init__.py":
        m = f[:-3]
        mod = __import__(".".join([__name__, m]), fromlist=[m])

        for k, v in mod.__dict__.items():
            if isfunction(v) and v.__module__ == mod.__name__ and not k.startswith("_"):
                func_dict[k] = magicdf(v)

__all__ = ["func_dict"]


# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:51 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : __init__.py.py
# @Desc   : The commons package's init file

from .parallel import job_factory
from .helpers import *

# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:52 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : __init__.py.py
# @Desc   : The techtable package's init file


from .funcs import func_dict
from .tt import TechTable

# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:54 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : tt.py
# @Desc   : The TechTable class

from lutar.commons import warnings
from lutar.techtable.funcs import func_dict

from functools import partial, update_wrapper
import pandas as pd


class TechTable:
    """


    """

    _df = None
    _raw_cols = None
    _verbose = None
    _alert_flag = {}

    def __init__(
        self, df=None, verbose=0, **kwargs,
    ):
        """

        * indexed with datetime is recommended


        """
        if df is not None:
            if isinstance(df, TechTable):
                df = df.df
            if isinstance(df, pd.DataFrame):
                if "datetime" in df.columns:
                    self._df = df.set_index("datetime")
                else:
                    self._df = df
        else:
            self._df = pd.DataFrame()

        self._raw_cols = self.df.columns.tolist()
        self._verbose = verbose
        for ti, func in func_dict.items():
            self.reg_func(ti_func=func, ti_name=ti)

    def reg_func(self, ti_func, ti_name=None):
        """
        ti_func must follow the standard format
        """
        ti_name = ti_name or ti_func.__name__
        super().__setattr__(
            ti_name.upper(), update_wrapper(partial(ti_func, self), ti_func)
        )

    @property
    def symbol_(self):
        return self.df.symbol.iat[0]

    @property
    def start_(self):
        return self.dt_index_[0]

    @property
    def end_(self):
        return self.dt_index_[-1]

    @property
    def dt_index_(self):
        return self.df.index

    @property
    def funcs_(self):
        return sorted([x for x in self.__dict__.keys() if x.isupper()])

    @property
    def df(self):
        return self._df

    @property
    def raw(self):
        return self.df.loc[:, self._raw_cols] if self._raw_cols else None

    def aligned(self, calendar=None, dfi=None):
        """align to a calendar, or according to the index

        * if align to dfi, the dates after self.df.dt_index_[-1] is ignored
        """
        if calendar is None:
            start_ts = max(self.dt_index_[0], dfi.dt_index_[0])
            end_ts = min(self.dt_index_[-1], dfi.dt_index_[-1])
            calendar = dfi.loc[start_ts:end_ts].index

        # 对齐
        earliest = self.df.index[0]
        calendar = calendar[calendar >= earliest]

        dfs_aligned = self.raw.reindex(index=calendar)
        dfs_aligned.symbol = dfs_aligned.symbol.fillna(self.symbol_)

        # 填充
        for col in ["volume", "turnover"]:
            try:
                dfs_aligned[col] = dfs_aligned[col].fillna(0)
            except KeyError:
                pass

        # use the last valid close to fill the subsequent nan
        dfs_aligned.close = dfs_aligned.close.fillna(method="ffill")
        for col in ["open", "high", "low"]:
            try:
                dfs_aligned[col] = dfs_aligned[col].fillna(dfs_aligned.close)
            except KeyError:
                pass

        # use the first valid open to fill all previous nan
        if dfs_aligned.index[0] > calendar[0]:
            dfs_aligned.open = dfs_aligned.open.fillna(method="bfill")
            for col in ["high", "low", "close"]:
                try:
                    dfs_aligned[col] = dfs_aligned[col].fillna(dfs_aligned.open)
                except KeyError:
                    pass
        return TechTable(dfs_aligned)

    def save(self, df_ti, cols=True):
        if isinstance(cols, bool) and cols:
            for c in df_ti.columns:
                c = str(c)
                if c not in self.raw.columns:
                    self.df.loc[df_ti.index, c] = df_ti[c]

        if isinstance(cols, str):
            self.df.loc[df_ti.index, cols] = df_ti[cols]
        if isinstance(cols, list):
            for c in cols:
                c = str(c)
                if c in df_ti.columns:
                    self.df.loc[df_ti.index, c] = df_ti[c]

    def __getattr__(self, item):
        # item不是TT的attribute，进入此函数
        if self.df is None:
            return None

        ITEM = item.upper()
        try:
            result = self.df.__getattr__(item)  # 优先尝试从self.df中读取现有数据
            if self._verbose > 0 and ITEM in self.__dict__.keys():
                if self._alert_flag.get(ITEM, True):
                    self._alert_flag[ITEM] = False  # only warn once
                    conflict_alert = (
                        f"The indicator function {item} is in conflict with internal df's attribute/column name. "
                        f"Returning TechTable.df.{item}.\n"
                        f"Use TechTable.{ITEM} to access the indicator function.\n"
                    )
                    warnings.warn(conflict_alert, UserWarning)
            return result
        except AttributeError:  # item 既不在TT的attribute中，又不在self.df的attribute中
            result = self.__dict__.get(ITEM, None)
            if result is None:
                indicator_alert = f"{item} is not in the internal df or the Indicator {ITEM} is not supported yet."
                raise KeyError(indicator_alert)

            # feature-generation function, return the function
            if ITEM.startswith("LHGEN_"):
                return result

            # 无需额外参数的ti，如ohlc4，第一次调用的时候自动计算
            # 尝试获取默认列名，如果不能获取，说明需要更多的参数，即应返回函数
            try:
                ti_col = result(name_only=True)
                s_ti = result(name_only=False)
                if isinstance(s_ti, pd.Series):  # normally should be series
                    self.df[ti_col] = s_ti
                return s_ti
            except:
                return result

    def __setattr__(self, key, value):
        if key in ["_raw_cols", "_df", "_area", "_alert_flag", "_verbose"]:
            super().__setattr__(key, value)

        elif key.upper() in self.__dict__.keys():
            msg = (
                f"Cannot set the '{key}' since it is an internal Technical Indicator. "
                f"Consider using tt['{key}'] to modify the column."
            )
            raise KeyError(msg)
        else:
            self._df.__setattr__(key, value)

    def __getitem__(self, item):
        return self.df.__getitem__(item)

    def __setitem__(self, key, value):
        return self.df.__setitem__(key, value)

    def __repr__(self):
        return self.df.__repr__()

    def __getstate__(self):
        return self.df, self._raw_cols, self._verbose, self._alert_flag

    def __setstate__(self, state):
        # self._df = state[0]
        # self._raw_cols = state[1]
        # self._verbose = state[2]
        # self._alert_flag = state[3]

        self._df, self._raw_cols, self._verbose, self._alert_flag = state

        for ti, func in func_dict.items():
            self.reg_func(ti_func=func, ti_name=ti)


# ----keep at the end



# -*- coding: utf-8 -*-
# @Time   : 2020-09-29 4:31 PM
# @Author : Zhiyi Lu
# @Email  : zhiyilv@outlook.com
# @File   : setup.py
# @Desc   : to install lutar


import setuptools
import io
import os


# Package meta-data.
NAME = "lutar"
DESCRIPTION = "Lu's Useful Technical Analysis Repository"
URL = "https://github.com/zhiyilv/lutar"
EMAIL = "zhiyilv@outlook.com"
AUTHOR = "Zhiyi Lu"
REQUIRES_PYTHON = ">=3.6.0"
VERSION = "0.1"

PACKAGES = setuptools.find_packages()
PACKAGES_DATA = {}

# What packages are required for this module to be executed?
REQUIRED = ["pandas", "joblib"]

# What packages are optional?
EXTRAS = {"serializing": [], "database": []}

CLASSIFIERS = [
    "Programming Language :: Python :: 3",
    "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
    "Operating System :: OS Independent",
    "Development Status :: 2 - Pre-Alpha",
]

here = os.path.abspath(os.path.dirname(__file__))

# Import the README and use it as the long-description.
# Note: this will only work if 'README.md' is present in your MANIFEST.in file!
try:
    with io.open(os.path.join(here, "README.md"), encoding="utf-8") as f:
        long_description = "\n" + f.read()
except FileNotFoundError:
    long_description = DESCRIPTION

setuptools.setup(
    name=NAME,
    version=VERSION,
    description=DESCRIPTION,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=AUTHOR,
    author_email=EMAIL,
    python_requires=REQUIRES_PYTHON,
    url=URL,
    packages=PACKAGES,
    install_requires=REQUIRED,
    zip_safe=False,
    package_data=PACKAGES_DATA,
)
